# Copyright 2024 MIT Probabilistic Computing Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import treescope


def pretty():
    treescope.register_as_default()
    treescope.register_autovisualize_magic()

    # Optional: enables automatic array visualization
    treescope.active_autovisualizer.set_interactive(treescope.ArrayAutovisualizer())


__all__ = [
    "pretty",
]
